@@ -87,6 +87,7 @@ from hypothesis.searchstrategy.collections import (
     ListStrategy,
     TupleStrategy,
     UniqueListStrategy,
+    UniqueSampledListStrategy,
 )
 from hypothesis.searchstrategy.datetime import (
     DateStrategy,
@@ -182,18 +183,17 @@ def cacheable(fn):
     # type: (T) -> T
     @proxies(fn)
     def cached_strategy(*args, **kwargs):
-        kwargs_cache_key = set()
         try:
-            for k, v in kwargs.items():
-                kwargs_cache_key.add((k, convert_value(v)))
+            kwargs_cache_key = {(k, convert_value(v)) for k, v in kwargs.items()}
         except TypeError:
             return fn(*args, **kwargs)
         cache_key = (fn, tuple(map(convert_value, args)), frozenset(kwargs_cache_key))
         try:
-            return STRATEGY_CACHE[cache_key]
+            if cache_key in STRATEGY_CACHE:
+                return STRATEGY_CACHE[cache_key]
         except TypeError:
             return fn(*args, **kwargs)
-        except KeyError:
+        else:
             result = fn(*args, **kwargs)
             if not isinstance(result, SearchStrategy) or result.is_cacheable:
                 STRATEGY_CACHE[cache_key] = result
@@ -776,6 +776,26 @@ def lists(
         for i, f in enumerate(unique_by):
             if not callable(f):
                 raise InvalidArgument("unique_by[%i]=%r is not a callable" % (i, f))
+        # Note that lazy strategies automatically unwrap when passed to a defines_strategy
+        # function.
+        if isinstance(elements, SampledFromStrategy):
+            element_count = len(elements.elements)
+            if min_size > element_count:
+                raise InvalidArgument(
+                    "Cannot create a collection of min_size=%r unique elements with "
+                    "values drawn from only %d distinct elements"
+                    % (min_size, element_count)
+                )
+
+            if max_size is not None:
+                max_size = min(max_size, element_count)
+            else:
+                max_size = element_count
+
+            return UniqueSampledListStrategy(
+                elements=elements, max_size=max_size, min_size=min_size, keys=unique_by
+            )
+
         return UniqueListStrategy(
             elements=elements, max_size=max_size, min_size=min_size, keys=unique_by
         )
@@ -1236,7 +1256,7 @@ def builds(
             "builds(), but is only allowed as a keyword arg"
         )
     required = required_args(target, args, kwargs) or set()
-    to_infer = set(k for k, v in kwargs.items() if v is infer)
+    to_infer = {k for k, v in kwargs.items() if v is infer}
     if required or to_infer:
         if isclass(target) and attr.has(target):
             # Use our custom introspection for attrs classes
@@ -1319,8 +1339,8 @@ def from_type(thing):
 
     For example, ``everything_except(int)`` returns a strategy that can
     generate anything that ``from_type()`` can ever generate, except for
-    instances of :class:python:int, and excluding instances of types
-    added via :func:~hypothesis.strategies.register_type_strategy.
+    instances of :class:`python:int`, and excluding instances of types
+    added via :func:`~hypothesis.strategies.register_type_strategy`.
 
     This is useful when writing tests which check that invalid input is
     rejected in a certain way.
