@@ -26,17 +26,19 @@ import hypothesis.internal.conjecture.utils as cu
 from hypothesis import Verbosity
 from hypothesis._settings import note_deprecation
 from hypothesis.errors import InvalidArgument
-from hypothesis.internal.compat import hrange, integer_types, text_type
+from hypothesis.internal.compat import hrange, integer_types
 from hypothesis.internal.coverage import check_function
 from hypothesis.internal.reflection import proxies
-from hypothesis.internal.validation import check_type
+from hypothesis.internal.validation import check_type, check_valid_interval
 from hypothesis.reporting import current_verbosity
 from hypothesis.searchstrategy import SearchStrategy
 
 if False:
-    from typing import Any, Union, Sequence, Tuple  # noqa
+    from typing import Any, Union, Sequence, Tuple, Optional  # noqa
     from hypothesis.searchstrategy.strategies import T  # noqa
 
+    Shape = Tuple[int, ...]  # noqa
+
 TIME_RESOLUTIONS = tuple("Y  M  D  h  m  s  ms  us  ns  ps  fs  as".split())
 
 
@@ -126,16 +128,6 @@ class ArrayStrategy(SearchStrategy):
     def __init__(self, element_strategy, shape, dtype, fill, unique):
         self.shape = tuple(shape)
         self.fill = fill
-        check_argument(
-            shape,
-            "Array shape must have at least one dimension, provided shape was {}",
-            shape,
-        )
-        check_argument(
-            all(isinstance(s, integer_types) for s in shape),
-            "Array shape must be integer in each dimension, provided shape was {}",
-            shape,
-        )
         self.array_size = int(np.prod(shape))
         self.dtype = dtype
         self.element_strategy = element_strategy
@@ -302,11 +294,10 @@ def fill_for(elements, unique, fill, name=""):
     return fill
 
 
-@st.composite
+@st.defines_strategy
 def arrays(
-    draw,  # type: Any
     dtype,  # type: Any
-    shape,  # type: Union[int, Sequence[int]]
+    shape,  # type: Union[int, Shape, st.SearchStrategy[Shape]]
     elements=None,  # type: st.SearchStrategy[Any]
     fill=None,  # type: st.SearchStrategy[Any]
     unique=False,  # type: bool
@@ -314,12 +305,11 @@ def arrays(
     # type: (...) -> st.SearchStrategy[np.ndarray]
     r"""Returns a strategy for generating :class:`numpy:numpy.ndarray`\ s.
 
-    * ``dtype`` may be any valid input to
-      :class:`numpy.dtype <numpy:numpy.dtype>` (this includes
-      :class:`~numpy:numpy.dtype` objects), or a strategy that generates such
-      values.
-    * ``shape`` may be an integer >= 0, a tuple of length >= 0 of such
-      integers, or a strategy that generates such values.
+    * ``dtype`` may be any valid input to :class:`~numpy:numpy.dtype`
+      (this includes :class:`~numpy:numpy.dtype` objects), or a strategy that
+      generates such values.
+    * ``shape`` may be an integer >= 0, a tuple of such integers, or a
+      strategy that generates such values.
     * ``elements`` is a strategy for generating values to put in the array.
       If it is None a suitable value will be inferred based on the dtype,
       which may give any legal value (including eg ``NaN`` for floats).
@@ -382,29 +372,60 @@ def arrays(
     hundreds or more elements, having a fill value is essential if you want
     your tests to run in reasonable time.
     """
+    # We support passing strategies as arguments for convenience, or at least
+    # for legacy reasons, but don't want to pay the perf cost of a composite
+    # strategy (i.e. repeated argument handling and validation) when it's not
+    # needed.  So we get the best of both worlds by recursing with flatmap,
+    # but only when it's actually needed.
     if isinstance(dtype, SearchStrategy):
-        dtype = draw(dtype)
+        return dtype.flatmap(
+            lambda d: arrays(d, shape, elements=elements, fill=fill, unique=unique)
+        )
+    if isinstance(shape, SearchStrategy):
+        return shape.flatmap(
+            lambda s: arrays(dtype, s, elements=elements, fill=fill, unique=unique)
+        )
+    # From here on, we're only dealing with values and it's relatively simple.
     dtype = np.dtype(dtype)
     if elements is None:
         elements = from_dtype(dtype)
-    if isinstance(shape, SearchStrategy):
-        shape = draw(shape)
-    if isinstance(shape, int):
+    if isinstance(shape, integer_types):
         shape = (shape,)
     shape = tuple(shape)
-    if not shape:
-        if dtype.kind != u"O":
-            return draw(elements)
+    check_argument(
+        all(isinstance(s, integer_types) for s in shape),
+        "Array shape must be integer in each dimension, provided shape was {}",
+        shape,
+    )
     fill = fill_for(elements=elements, unique=unique, fill=fill)
-    return draw(ArrayStrategy(elements, shape, dtype, fill, unique))
+    return ArrayStrategy(elements, shape, dtype, fill, unique)
 
 
 @st.defines_strategy
-def array_shapes(min_dims=1, max_dims=3, min_side=1, max_side=10):
-    # type: (int, int, int, int) -> st.SearchStrategy[Tuple[int, ...]]
+def array_shapes(min_dims=1, max_dims=None, min_side=1, max_side=None):
+    # type: (int, int, int, int) -> st.SearchStrategy[Shape]
     """Return a strategy for array shapes (tuples of int >= 1)."""
-    order_check("dims", 1, min_dims, max_dims)
-    order_check("side", 1, min_side, max_side)
+    check_type(integer_types, min_dims, "min_dims")
+    check_type(integer_types, min_side, "min_side")
+    if min_dims > 32:
+        raise InvalidArgument(
+            "Got min_dims=%r, but numpy does not support arrays greater than 32 dimensions"
+            % min_dims
+        )
+    if max_dims is None:
+        max_dims = min(min_dims + 2, 32)
+    check_type(integer_types, max_dims, "max_dims")
+    if max_dims > 32:
+        raise InvalidArgument(
+            "Got max_dims=%r, but numpy does not support arrays greater than 32 dimensions"
+            % max_dims
+        )
+    if max_side is None:
+        max_side = min_side + 5
+    check_type(integer_types, max_side, "max_side")
+    order_check("dims", 0, min_dims, max_dims)
+    order_check("side", 0, min_side, max_side)
+
     return st.lists(
         st.integers(min_side, max_side), min_size=min_dims, max_size=max_dims
     ).map(tuple)
@@ -430,7 +451,20 @@ def defines_dtype_strategy(strat):
     @st.defines_strategy
     @proxies(strat)
     def inner(*args, **kwargs):
-        return strat(*args, **kwargs).map(np.dtype)
+        strategy = strat(*args, **kwargs)
+
+        def convert_to_dtype(x):
+            """Helper to debug issue #1798."""
+            try:
+                return np.dtype(x)
+            except ValueError:
+                print(
+                    "Got invalid dtype value=%r from strategy=%r, function=%r"
+                    % (x, strategy, strat)
+                )
+                raise
+
+        return strategy.map(convert_to_dtype)
 
     return inner
 
@@ -461,7 +495,7 @@ def dtype_factory(kind, sizes, valid_sizes, endianness):
             valid_sizes,
         )
         if all(isinstance(s, int) for s in sizes):
-            sizes = sorted(set(s // 8 for s in sizes))
+            sizes = sorted({s // 8 for s in sizes})
     strat = st.sampled_from(sizes)
     if "{}" not in kind:
         kind += "{}"
@@ -604,9 +638,7 @@ def array_dtypes(
     """Return a strategy for generating array (compound) dtypes, with members
     drawn from the given subtype strategy."""
     order_check("size", 0, min_size, max_size)
-    native_strings = st.text()  # type: SearchStrategy[Any]
-    if text_type is not str:  # pragma: no cover
-        native_strings = st.binary()
+    native_strings = st.from_type(str).filter(bool)  # See issue #1798 re: filter!
     elements = st.tuples(native_strings, subtype_strategy)
     if allow_subarrays:
         elements |= st.tuples(
@@ -638,3 +670,246 @@ def nested_dtypes(
     return st.recursive(
         subtype_strategy, lambda x: array_dtypes(x, allow_subarrays=True), max_leaves
     ).filter(lambda d: max_itemsize is None or d.itemsize <= max_itemsize)
+
+
+@st.defines_strategy
+def valid_tuple_axes(ndim, min_size=0, max_size=None):
+    # type: (int, int, int) -> st.SearchStrategy[Shape]
+    """Return a strategy for generating permissible tuple-values for the
+    ``axis`` argument for a numpy sequential function (e.g.
+    :func:`numpy:numpy.sum`), given an array of the specified
+    dimensionality.
+
+    All tuples will have an length >= min_size and <= max_size. The default
+    value for max_size is ``ndim``.
+
+    Examples from this strategy shrink towards an empty tuple, which render
+    most sequential functions as no-ops.
+
+    The following are some examples drawn from this strategy.
+
+    .. code-block:: pycon
+
+        >>> [valid_tuple_axes(3).example() for i in range(4)]
+        [(-3, 1), (0, 1, -1), (0, 2), (0, -2, 2)]
+
+    ``valid_tuple_axes`` can be joined with other strategies to generate
+    any type of valid axis object, i.e. integers, tuples, and ``None``:
+
+    .. code-block:: pycon
+
+        any_axis_strategy = none() | integers(-ndim, ndim - 1) | valid_tuple_axes(ndim)
+
+    """
+    if max_size is None:
+        max_size = ndim
+
+    check_type(integer_types, ndim, "ndim")
+    check_type(integer_types, min_size, "min_size")
+    check_type(integer_types, max_size, "max_size")
+    order_check("size", 0, min_size, max_size)
+    check_valid_interval(max_size, ndim, "max_size", "ndim")
+
+    # shrink axis values from negative to positive
+    axes = st.integers(0, max(0, 2 * ndim - 1)).map(
+        lambda x: x if x < ndim else x - 2 * ndim
+    )
+    return st.lists(axes, min_size, max_size, unique_by=lambda x: x % ndim).map(tuple)
+
+
+class BroadcastShapeStrategy(SearchStrategy):
+    def __init__(self, shape, min_dims, max_dims, min_side, max_side):
+        assert 0 <= min_side <= max_side
+        assert 0 <= min_dims <= max_dims <= 32
+        SearchStrategy.__init__(self)
+        self.shape = shape
+        self.side_strat = st.integers(min_side, max_side)
+        self.min_dims = min_dims
+        self.max_dims = max_dims
+        self.min_side = min_side
+        self.max_side = max_side
+
+    def do_draw(self, data):
+        elements = cu.many(
+            data,
+            min_size=self.min_dims,
+            max_size=self.max_dims,
+            average_size=min(
+                max(self.min_dims * 2, self.min_dims + 5),
+                0.5 * (self.min_dims + self.max_dims),
+            ),
+        )
+        result = []
+        reversed_shape = tuple(self.shape[::-1])
+        while elements.more():
+            if len(result) < len(self.shape):
+                # Shrinks towards original shape
+                if reversed_shape[len(result)] == 1:
+                    if self.min_side <= 1 and not data.draw(st.booleans()):
+                        side = 1
+                    else:
+                        side = data.draw(self.side_strat)
+                elif self.max_side >= reversed_shape[len(result)] and (
+                    not self.min_side <= 1 <= self.max_side or data.draw(st.booleans())
+                ):
+                    side = reversed_shape[len(result)]
+                else:
+                    side = 1
+            else:
+                side = data.draw(self.side_strat)
+            result.append(side)
+        assert self.min_dims <= len(result) <= self.max_dims
+        assert all(self.min_side <= s <= self.max_side for s in result)
+        return tuple(reversed(result))
+
+
+@st.defines_strategy
+def broadcastable_shapes(shape, min_dims=0, max_dims=None, min_side=1, max_side=None):
+    # type: (Shape, int, int, int, int) -> st.SearchStrategy[Shape]
+    """Return a strategy for generating shapes that are broadcast-compatible
+    with the provided shape.
+
+    Examples from this strategy shrink towards a shape with length ``min_dims``.
+    The size of an aligned dimension shrinks towards being a singleton. The
+    size of an unaligned dimension shrink towards ``min_side``.
+
+    * ``shape`` a tuple of integers
+    * ``min_dims`` The smallest length that the generated shape can possess.
+    * ``max_dims`` The largest length that the generated shape can possess.
+      shape can possess. Cannot exceed 32. The default-value for ``max_dims``
+      is ``2 + max(len(shape), min_dims)``.
+    * ``min_side`` The smallest size that an unaligned dimension can possess.
+    * ``max_side`` The largest size that an unaligned dimension can possess.
+      The default value is 2 + 'size-of-largest-aligned-dimension'.
+
+    The following are some examples drawn from this strategy.
+
+    .. code-block:: pycon
+
+        >>> [broadcastable_shapes(shape=(2, 3)).example() for i in range(5)]
+        [(1, 3), (), (2, 3), (2, 1), (4, 1, 3), (3, )]
+
+    """
+    check_type(tuple, shape, "shape")
+    strict_check = max_side is None or max_dims is None
+    check_type(integer_types, min_side, "min_side")
+    check_type(integer_types, min_dims, "min_dims")
+
+    if max_dims is None:
+        max_dims = max(len(shape), min_dims) + 2
+    else:
+        check_type(integer_types, max_dims, "max_dims")
+
+    if max_side is None:
+        max_side = max(tuple(shape[-max_dims:]) + (min_side,)) + 2
+    else:
+        check_type(integer_types, max_side, "max_side")
+
+    order_check("dims", 0, min_dims, max_dims)
+    order_check("side", 0, min_side, max_side)
+
+    if 32 < max_dims:
+        raise InvalidArgument("max_dims cannot exceed 32")
+
+    dims, bnd_name = (max_dims, "max_dims") if strict_check else (min_dims, "min_dims")
+
+    # check for unsatisfiable min_side
+    if not all(min_side <= s for s in shape[::-1][:dims] if s != 1):
+        raise InvalidArgument(
+            "Given shape=%r, there are no broadcast-compatible "
+            "shapes that satisfy: %s=%s and min_side=%s"
+            % (shape, bnd_name, dims, min_side)
+        )
+
+    # check for unsatisfiable [min_side, max_side]
+    if not (
+        min_side <= 1 <= max_side or all(s <= max_side for s in shape[::-1][:dims])
+    ):
+        raise InvalidArgument(
+            "Given shape=%r, there are no broadcast-compatible shapes "
+            "that satisfy: %s=%s and [min_side=%s, max_side=%s]"
+            % (shape, bnd_name, dims, min_side, max_side)
+        )
+
+    if not strict_check:
+        # reduce max_dims to exclude unsatisfiable dimensions
+        for n, s in zip(range(max_dims), reversed(shape)):
+            if s < min_side and s != 1:
+                max_dims = n
+                break
+            elif not (min_side <= 1 <= max_side or s <= max_side):
+                max_dims = n
+                break
+
+    return BroadcastShapeStrategy(
+        shape,
+        min_dims=min_dims,
+        max_dims=max_dims,
+        min_side=min_side,
+        max_side=max_side,
+    )
+
+
+@st.defines_strategy
+def integer_array_indices(shape, result_shape=array_shapes(), dtype="int"):
+    # type: (Shape, SearchStrategy[Shape], np.dtype) -> st.SearchStrategy[Tuple[np.ndarray, ...]]
+    """Return a search strategy for tuples of integer-arrays that, when used
+    to index into an array of shape ``shape``, given an array whose shape
+    was drawn from ``result_shape``.
+
+    Examples from this strategy shrink towards the tuple of index-arrays::
+
+        len(shape) * (np.zeros(drawn_result_shape, dtype), )
+
+    * ``shape`` a tuple of integers that indicates the shape of the array,
+      whose indices are being generated.
+    * ``result_shape`` a strategy for generating tuples of integers, which
+      describe the shape of the resulting index arrays. The default is
+      :func:`~hypothesis.extra.numpy.array_shapes`.  The shape drawn from
+      this strategy determines the shape of the array that will be produced
+      when the corresponding example from ``integer_array_indices`` is used
+      as an index.
+    * ``dtype`` the integer data type of the generated index-arrays. Negative
+      integer indices can be generated if a signed integer type is specified.
+
+    Recall that an array can be indexed using a tuple of integer-arrays to
+    access its members in an arbitrary order, producing an array with an
+    arbitrary shape. For example:
+
+    .. code-block:: pycon
+
+        >>> from numpy import array
+        >>> x = array([-0, -1, -2, -3, -4])
+        >>> ind = (array([[4, 0], [0, 1]]),)  # a tuple containing a 2D integer-array
+        >>> x[ind]  # the resulting array is commensurate with the indexing array(s)
+        array([[-4,  0],
+               [0, -1]])
+
+    Note that this strategy does not accommodate all variations of so-called
+    'advanced indexing', as prescribed by NumPy's nomenclature.  Combinations
+    of basic and advanced indexes are too complex to usefully define in a
+    standard strategy; we leave application-specific strategies to the user.
+    Advanced-boolean indexing can be defined as ``arrays(shape=..., dtype=bool)``,
+    and is similarly left to the user.
+    """
+    check_type(tuple, shape, "shape")
+    check_argument(
+        shape and all(isinstance(x, integer_types) and x > 0 for x in shape),
+        "shape=%r must be a non-empty tuple of integers > 0" % (shape,),
+    )
+    check_type(SearchStrategy, result_shape, "result_shape")
+    check_argument(
+        np.issubdtype(dtype, np.integer), "dtype=%r must be an integer dtype" % (dtype,)
+    )
+    signed = np.issubdtype(dtype, np.signedinteger)
+
+    def array_for(index_shape, size):
+        return arrays(
+            dtype=dtype,
+            shape=index_shape,
+            elements=st.integers(-size if signed else 0, size - 1),
+        )
+
+    return result_shape.flatmap(
+        lambda index_shape: st.tuples(*[array_for(index_shape, size) for size in shape])
+    )
